import numbers
import re

from dataclasses import dataclass
from typing import Any, Callable, Dict, NoReturn, Optional, Union
from ..utils import raise_error
from ..definitions import FieldID, FieldName, BasicDataclass, TypeDefinition, GenFieldDefinition
# TODO: add DEFAULT to dataclasses


def echo(val: Any) -> Any:
    return val


# Symbol Table Field Definition fields
@dataclass
class SymbolTableFieldDefinition(BasicDataclass):
    Def: GenFieldDefinition     # 0: JADN field definition
    Opt: dict                   # 1: Field Options (dict format)
    cTag: Optional[int]         # 2: Field containing external choice tag (tagid option)


# Symbol Table fields
@dataclass
class SymbolTableField(BasicDataclass):
    TypeDef: TypeDefinition                                     # 0: JADN type definition
    Encode: Callable[['SymbolTableField', Any, 'Codec'], Any]   # 1: Encoder for this type
    Decode: Callable[['SymbolTableField', Any, 'Codec'], Any]   # 2: Decoder for this type
    EncType: type                                               # 3: Encoded value type
    TypeOpts: dict                                              # 4: Type Options (dict format)
    # 5: Format semantic validation - returns True if valid
    FormatValidate: Callable[[Any], Any] = echo
    # 6: Format encode conversion - returns serialized data representation
    FormatEncode: Callable[[Any], Any] = echo
    # 7: Format decode conversion - returns API value
    FormatDecode: Callable[[Any], Any] = echo
    # 8: Decode: Encoded field key or enum value to API
    dMap: Dict[Union[int, str], Union[int, str]] = None
    # 9: Decode: Encode: API field key or enum value to Encoded
    eMap: Dict[Union[int, str], Union[int, str]] = None
    # 10: Field entries (definition and decoded options)
    Fld: Dict[str, SymbolTableFieldDefinition] = None


# Codec Table fields
@dataclass
class CodecTableField(BasicDataclass):
    # 0: Decode function
    Dec: Callable[[SymbolTableField, Any, 'Codec'], Any] = None
    # 1: Encode function
    Enc: Callable[[SymbolTableField, Any, 'Codec'], Any] = None
    # 2: Encoded type
    eType: type = None


def _bad_index(ts: SymbolTableField, k: int, val: list) -> NoReturn:
    td = ts.TypeDef
    raise_error(f'{td.TypeName}({td.BaseType}): array index {k} out of bounds ({len(ts.Fld)}, {len(val)})')


def _bad_choice(ts: SymbolTableField, val: Any) -> NoReturn:
    td = ts.TypeDef
    raise_error(f'{td.TypeName}: choice must have one value: {val}')


def _bad_value(ts: SymbolTableField, val: Any, fld: GenFieldDefinition = None) -> NoReturn:
    td = ts.TypeDef
    if fld is not None:
        raise_error(f'{td.TypeName}({td.BaseType}): missing required field "{fld.FieldName}": {val}')
    else:
        v = next(iter(val)) if isinstance(val, dict) else val
        raise_error(f'{td.TypeName}({td.BaseType}): bad value: {v}')


# fail forces rejection of boolean vals for number types
def _check_type(ts: SymbolTableField, val: Any, vtype: type, fail=False) -> NoReturn:
    if vtype is not None:
        if fail or not isinstance(val, vtype):
            td = ts.TypeDef
            tn = f"{td.TypeName}({td.BaseType if td else 'Primitive'})"
            raise_error(f'{tn}: {val} is not {vtype}')


def _format_encode(ts: SymbolTableField, val: Any) -> Any:
    try:
        ts.FormatValidate(val)
    except ValueError:
        raise_error(f'{ts.TypeDef.TypeName}: {val} is not format "{ts.TypeOpts["format"]}"')
    return ts.FormatEncode(val)


def _format_decode(ts: SymbolTableField, val: Any) -> Any:
    aval = ts.FormatDecode(val)
    try:
        ts.FormatValidate(aval)
    except ValueError:
        raise_error(f'{ts.TypeDef.TypeName}: {val} is not format "{ts.TypeOpts["format"]}"')
    return aval


def _check_key(ts: SymbolTableField, val):
    try:
        return int(val) if isinstance(next(iter(ts.dMap)), int) else val
    except ValueError:
        raise_error(f'{ts.TypeDef.TypeName}: {val} is not a valid field ID')


def _check_pattern(ts: SymbolTableField, val):
    op = ts.TypeOpts
    if 'pattern' in op and not re.match(op['pattern'], val):
        tn = ts.TypeDef.TypeName
        raise_error(f'{tn}: string "{val}" does not match {op["pattern"]}')
    return val


def _check_range(ts: SymbolTableField, val):
    op = ts.TypeOpts
    tn = ts.TypeDef.TypeName
    if 'minv' in op and val < op['minv']:
        raise_error(f'{tn}: {val} < minimum {op["minv"]}')
    if 'maxv' in op and val > op['maxv']:
        raise_error(f'{tn}: {val} < maximum {op["maxv"]}')
    return val


def _check_frange(ts: SymbolTableField, val):
    op = ts.TypeOpts
    tn = ts.TypeDef.TypeName
    if 'minf' in op and val < op['minf']:
        raise_error(f'{tn}: {val} < minimum {op["minf"]}')
    if 'maxf' in op and val > op['maxf']:
        raise_error(f'{tn}: {val} < maximum {op["maxf"]}')
    return val


def _check_size(ts: SymbolTableField, val):
    op = ts.TypeOpts
    tn = ts.TypeDef.TypeName
    if 'minv' in op and len(val) < op['minv']:
        raise_error(f'{tn}: length {len(val)} < minimum {op["minv"]}')
    if 'maxv' in op and len(val) > op['maxv']:
        raise_error(f'{tn}: length {len(val)} > maximum {op["maxv"]}')
    return val


def _extra_value(ts: SymbolTableField, val, extra: set):
    td = ts.TypeDef
    raise_error(f'{td.TypeName}({td.BaseType}): unexpected field: \"{", ".join(str(k) for k in extra)}\"')


def _encode_binary(ts: SymbolTableField, aval, codec: 'Codec'):    # Encode bytes to string
    _check_type(ts, aval, bytes)
    _check_size(ts, aval)
    return _format_encode(ts, aval)


def _decode_binary(ts: SymbolTableField, sval, codec: 'Codec'):    # Decode ASCII string to bytes
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, bytes)        # assert format decode returns correct type
    return _check_size(ts, aval)


def _encode_boolean(ts: SymbolTableField, val, codec: 'Codec'):
    _check_type(ts, val, bool)
    return val


def _decode_boolean(ts: SymbolTableField, val, codec: 'Codec'):
    _check_type(ts, val, bool)
    return val


def _encode_integer(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, numbers.Integral, isinstance(aval, bool))
    _check_range(ts, aval)
    return _format_encode(ts, aval)


def _decode_integer(ts: SymbolTableField, sval, codec: 'Codec'):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, numbers.Integral, isinstance(aval, bool))
    return _check_range(ts, aval)


def _encode_number(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, numbers.Real, isinstance(aval, bool))
    _check_frange(ts, aval)
    return _format_encode(ts, aval)


def _decode_number(ts: SymbolTableField, sval, codec: 'Codec'):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, numbers.Real, isinstance(aval, bool))
    return _check_range(ts, aval)


def _encode_null(ts: SymbolTableField, aval, codec: 'Codec'):
    if aval:                            # Treat any false-y value as Null: None, False, 0, '', [], set(), {}
        _bad_value(ts, aval)
    return aval


def _decode_null(ts: SymbolTableField, val, codec: 'Codec'):
    if val:
        _bad_value(ts, val)
    return val


def _encode_string(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, type(''))
    _check_size(ts, aval)
    _check_pattern(ts, aval)
    return _format_encode(ts, aval)


def _decode_string(ts: SymbolTableField, sval, codec: 'Codec'):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, type(''))
    _check_size(ts, aval)
    return _check_pattern(ts, aval)


def _encode_enumerated(ts: SymbolTableField, aval, codec: 'Codec'):  # pylint: disable=R1710
    # TODO: Serialization
    _check_type(ts, aval, type(next(iter(ts.eMap))))
    if aval in ts.eMap:
        return ts.eMap[aval]
    td = ts.TypeDef
    raise_error(f'{td.BaseType}: {aval} is not a valid {td.TypeName}')


def _decode_enumerated(ts: SymbolTableField, sval, codec: 'Codec'):  # pylint: disable=R1710
    _check_type(ts, sval, type(next(iter(ts.dMap))))
    if sval in ts.dMap:
        return ts.dMap[sval]
    td = ts.TypeDef
    raise_error(f'{td.BaseType}: {sval} is not a valid {td.TypeName}')


def _encode_choice(ts: SymbolTableField, val, codec: 'Codec'):
    _check_type(ts, val, dict)
    if len(val) != 1:
        _bad_choice(ts, val)
    k, v = next(iter(val.items()))
    if k not in ts.eMap:
        _bad_value(ts, val)
    k = ts.eMap[k]
    f = ts.Fld[k].Def
    return {k: codec.encode(f.FieldType, v)}


def _decode_choice(ts: SymbolTableField, val, codec: 'Codec'):  # Map Choice:  val == {key: value}
    _check_type(ts, val, dict)
    if len(val) != 1:
        _bad_choice(ts, val)
    k, v = next(iter(val.items()))
    k = _check_key(ts, k)
    if k not in ts.dMap:
        _bad_value(ts, val)
    f = ts.Fld[k].Def
    k = ts.dMap[k]
    return {k: codec.decode(f.FieldType, v)}


def _encode_maprec(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, dict)
    sval = ts.EncType()
    assert isinstance(sval, (list, dict))
    fx = FieldName if codec.verbose_str else FieldID  # Verbose or minified identifier strings
    fnames = [f.Def.FieldName for f in ts.Fld.values()]
    for f in ts.TypeDef.Fields:
        fs = ts.Fld[f[fx]]  # Symtab entry for field
        fd = fs.Def  # JADN field definition from symtab
        fname = fd[FieldName]  # Field name
        fopts = fs.Opt  # Field options dict
        ctag = fs.cTag
        if ctag is not None:  # Type of this field is specified by contents of another field
            e = codec.encode(fd.FieldType, {aval[ctag]: aval[fname]})
            sv = next(iter(e.values()))
        else:
            sv = codec.encode(fd.FieldType, aval[fname]) if fname in aval else None
        if sv is None and ('minc' not in fopts or fopts['minc'] > 0):  # Missing required field
            _bad_value(ts, aval, fd)
        if isinstance(sval, list):  # Concise Record
            sval.append(sv)
        elif sv is not None:  # Map or Verbose Record
            sval[fd[fx]] = sv

    if set(aval) - set(fnames):
        _extra_value(ts, aval, fnames)
    if isinstance(sval, list):
        while sval and sval[-1] is None:  # Strip non-populated trailing optional values
            sval.pop()
    return sval


def _decode_maprec(ts: SymbolTableField, sval, codec: 'Codec'):
    _check_type(ts, sval, ts.EncType)
    val = sval
    if ts.EncType == dict:
        val = {_check_key(ts, k): v for k, v in sval.items()}
    aval = dict()
    fx = FieldName if codec.verbose_str else FieldID  # Verbose or minified identifier strings
    fnames = list(ts.Fld)
    for f in ts.TypeDef.Fields:
        fs = ts.Fld[f[fx]]  # Symtab entry for field
        fd = fs.Def  # JADN field definition from symtab
        fopts = fs.Opt  # Field options dict
        if isinstance(val, dict):
            fn = f[fx]
            sv = val[fn] if fn in val else None
        else:
            fn = fd[FieldID] - 1
            sv = val[fn] if len(val) > fn else None
        if sv is not None:
            ctag = fs.cTag
            if ctag is not None:  # Type of this field is specified by contents of another field
                ct = ctag if isinstance(val, dict) else ts.eMap[ctag] - 1
                av = codec.decode(fd.FieldType, {sval[ct]: sv})
                aval[fd[FieldName]] = next(iter(av.values()))
            else:
                aval[fd[FieldName]] = codec.decode(fd.FieldType, sv)
        else:
            if 'minc' not in fopts or fopts['minc'] > 0:
                _bad_value(ts, val, fd)
    extra = set(val) - set(fnames) if isinstance(val, dict) else set(val[len(ts.Fld):])
    if extra:
        _extra_value(ts, val, extra)
    return aval


def _encode_array(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, list)
    sval = list()
    if len(aval) > len(ts.Fld):
        _extra_value(ts, aval, set(aval[len(ts.Fld):]))
    for fn in ts.TypeDef.Fields:
        f = ts.Fld[fn.FieldID].Def  # Use symtab field definition
        fx = f.FieldID - 1
        fopts = ts.Fld[fx + 1].Opt
        av = aval[fx] if len(aval) > fx else None
        if av is not None:
            if 'tagid' in fopts:
                choice_type = aval[int(fopts['tagid']) - 1]
                e = codec.encode(f.FieldType, {choice_type: av})
                sv = e[next(iter(e))]
            else:
                sv = codec.encode(f.FieldType, av)
            sval.append(sv)
        else:
            sval.append(None)
            if 'minc' in fopts and fopts['minc'] > 0:   # Value is required
                _bad_value(ts, aval, f)
    while sval and sval[-1] is None:            # Strip non-populated trailing optional values
        sval.pop()
    return _format_encode(ts, sval)


def _decode_array(ts: SymbolTableField, sval, codec: 'Codec'):  # Ordered list of types, returned as a list
    val = _format_decode(ts, sval)
    _check_type(ts, val, list)
    aval = list()
    if len(val) > len(ts.Fld):
        _extra_value(ts, val, set(aval[len(ts.Fld):]))
    for fn in ts.TypeDef.Fields:
        f = ts.Fld[fn[FieldID]].Def  # Use symtab field definition
        fx = f[FieldID] - 1
        fopts = ts.Fld[fx + 1].Opt
        sv = val[fx] if len(val) > fx else None
        if sv is not None:
            if 'tagid' in fopts:
                choice_type = val[int(fopts['tagid']) - 1]
                d = codec.decode(f.FieldType, {choice_type: sv})  # TODO: fix str/int handling of choice
                av = d[next(iter(d))]
            else:
                av = codec.decode(f.FieldType, sv)
            aval.append(av)
        else:
            aval.append(None)
            if 'minc' not in fopts or fopts['minc'] > 0:
                _bad_value(ts, val, f)
    while aval and aval[-1] is None:  # Strip non-populated trailing optional values
        aval.pop()
    return aval


def _encode_array_of(ts: SymbolTableField, val, codec: 'Codec'):
    _check_type(ts, val, list)
    _check_size(ts, val)
    return [codec.encode(ts.TypeOpts['vtype'], v) for v in val]


def _decode_array_of(ts: SymbolTableField, val, codec: 'Codec'):
    _check_type(ts, val, list)
    _check_size(ts, val)
    return [codec.decode(ts.TypeOpts['vtype'], v) for v in val]


def _encode_map_of(ts: SymbolTableField, aval, codec: 'Codec'):
    _check_type(ts, aval, dict)
    _check_size(ts, aval)
    to = ts.TypeOpts
    return {codec.encode(to['ktype'], k): codec.encode(to['vtype'], v) for k, v in aval.items()}


def _decode_map_of(ts: SymbolTableField, sval, codec: 'Codec'):
    _check_type(ts, sval, dict)
    _check_size(ts, sval)
    return {k: codec.decode(ts.TypeOpts['vtype'], v) for k, v in sval.items()}


enctab: Dict[str, CodecTableField] = {  # decode, encode, min encoded type
    'Binary': CodecTableField(_decode_binary, _encode_binary, str),
    'Boolean': CodecTableField(_decode_boolean, _encode_boolean, bool),
    'Integer': CodecTableField(_decode_integer, _encode_integer, int),
    'Number': CodecTableField(_decode_number, _encode_number, float),
    'Null': CodecTableField(_decode_null, _encode_null, str),
    'String': CodecTableField(_decode_string, _encode_string, str),
    'Enumerated': CodecTableField(_decode_enumerated, _encode_enumerated, int),
    'Choice': CodecTableField(_decode_choice, _encode_choice, dict),
    'Array': CodecTableField(_decode_array, _encode_array, list),
    'ArrayOf': CodecTableField(_decode_array_of, _encode_array_of, list),
    'Map': CodecTableField(_decode_maprec, _encode_maprec, dict),
    'MapOf': CodecTableField(_decode_map_of, _encode_map_of, dict),
    'Record': CodecTableField(),  # Dynamic values
}



"""
Translate JADN to HTML or Markdown property tables
"""
from datetime import datetime
from enum import Enum
from typing import Dict, List, NoReturn, Type, Union
from ..definitions import (
    TypeName, BaseType, TypeOptions, TypeDesc, Fields, ItemID, ItemValue, ItemDesc, FieldID, INFO_ORDER, PRIMITIVE_TYPES,
    is_builtin
)
from ..utils import jadn2fielddef, jadn2typestr, topts_s2d


# --------- Basic Conversion Class -----------------
class BasicConversion:
    """
    Basic JADN -> FORMAT conversion class

    doc_begin - initial content
    doc_end - closing content, if any
    sect - section heading for human document formats, nothing for machine-readable schemas
    meta_begin - begin meta content
    meta_item - most meta items
    meta_imps - special handling for imports meta item
    meta_end - close meta content
    type_begin - begin type definition
    type_item - add field to type definition
    type_end - close type definition
    """
    doc: str
    schema: dict

    def __init__(self, schema: dict):
        self.doc = ''
        self.schema = schema
        # What happens here?

    def convert(self) -> str:
        self.doc = self.doc_begin()
        self.doc += self.write_meta()
        self.doc += self.write_types()
        self.doc += self.doc_end()
        return self.doc

    # Helper functions
    def write_meta(self) -> str:
        meta = ''
        if info := self.schema.get('info', None):
            meta += self.meta_begin()
            mlist = [k for k in INFO_ORDER if k in info]
            for h in mlist + list(set(info) - set(mlist)):
                mh = info[h]
                meta += self.meta_item(h, mh)
            meta += self.meta_end()
        return meta

    def write_types(self) -> str:
        types = ''
        for td in self.schema['types']:
            ttype = jadn2typestr(td[BaseType], td[TypeOptions])
            to = topts_s2d(td[TypeOptions])
            if not is_builtin(td[BaseType]):
                types += 'Error - bad type: ' + str(td) + '\n'
            elif td[BaseType] in (PRIMITIVE_TYPES + ('ArrayOf', 'MapOf')) or 'enum' in to or 'pointer' in to:
                cls = ['b', 's', 'd']
                types += self.type_begin('', '', ['Type Name', 'Type Definition', 'Description'], cls)
                types += self.type_item([td[TypeName], ttype, td[TypeDesc]], cls)
            elif td[BaseType] == 'Enumerated':
                cls = ['n', 'b', 'd']
                types += self._tbegin(to, td[TypeName], ttype, ['ID', 'Name', 'Description'], cls)
                for fd in td[Fields]:
                    types += self._titem(to, [str(fd[ItemID]), fd[ItemValue], fd[ItemDesc]], cls)
            else:  # Array, Choice, Map, Record
                cls = ['n', 'b', 's', 'n', 'd']
                if td[BaseType] == 'Array':
                    cls2 = ['n', 's', 'n', 'd']  # Don't print ".ID" in type name but display fields as compact
                    types += self._tbegin(to, td[TypeName], ttype, ['ID', 'Type', '#', 'Description'], cls2)
                    # to.update({'id': None})
                else:
                    types += self._tbegin(to, td[TypeName], ttype, ['ID', 'Name', 'Type', '#', 'Description'], cls)
                for fd in td[Fields]:
                    fn, ftype, fmult, fdesc = jadn2fielddef(fd, td)
                    types += self._titem(to, [str(fd[FieldID]), fn, ftype, fmult, fdesc], cls)
            types += self.type_end()
        return types

    # Override functions
    def doc_begin(self) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `doc_begin`')

    def doc_end(self) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `doc_end`')

    def sect(self, num, name) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `sect`')

    def meta_begin(self) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `meta_begin`')

    def meta_item(self, key: str, val: Union[dict, list, str]) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `meta_item`')

    def meta_end(self) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `meta_end`')

    def type_begin(self, tname: str, ttype: str, headers: List[str], cls: List[str]) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `type_begin`')

    def type_item(self, row: list, cls: list) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `type_item`')

    def type_end(self) -> str:
        raise NotImplementedError(f'{self.__class__.__name__} does not implement `type_end`')

    # Utility functions
    def _tbegin(self, to: dict, name: str, tdef: str, head: List[str], cls: List[str]) -> str:
        h = head
        c = cls
        if 'id' in to:
            h = [head[0]] + head[2:]
            c = [cls[0]] + cls[2:]
        return self.type_begin(name, tdef, h, c)

    def _titem(self, to: dict, fitems: List[str], cls: List[str]) -> str:
        f = fitems
        c = cls
        if 'id' in to:
            f = [fitems[0]] + fitems[2:]
            label = '**' + fitems[1] + '**::' if fitems[1] else ''
            f[-1] = label + f[-1]
            c = [cls[0]] + cls[2:]
        return self.type_item(f, c)


# --------- Markdown output -----------------
class MarkdownConversion(BasicConversion):
    def doc_begin(self) -> str:
        return '## Schema\n'

    def doc_end(self) -> str:
        return ''

    def sect(self, num, name) -> str:
        # n = '.'.join([str(n) for n in num]) + ' '
        # return '\n' + len(num)*'#' + ' ' + n + name + '\n'
        return ''

    def meta_begin(self) -> str:
        return '| . | . |\n| ---: | :--- |\n'

    def meta_item(self, key: str, val: Union[dict, list, str]) -> str:
        if key == 'exports':
            sval = ', '.join(val)
        elif key in ('imports', 'config'):
            sval = ' '.join(['**' + k + '**:&nbsp;' + str(v).replace('|', '&vert;') for k, v in val.items()])
        else:
            sval = val
        return f'| **{key}:** | {sval} |\n'

    def meta_end(self) -> str:
        return ''

    def type_begin(self, tname: str, ttype: str, headers: List[str], cls: List[str]) -> str:
        assert len(headers) == len(cls)
        ch = {'n': '---:', 'h': '---:', 's': ':---', 'd': ':---', 'b': ':---'}
        clh = [ch[c] if c in ch else '---' for c in cls]
        to = f' ({ttype})' if ttype else ''
        tc = f'\n**_Type: {tname}{to}_**' if tname else ''
        return f"{tc}\n\n| {' | '.join(headers)} |\n| {' | '.join(clh)} |\n"

    def type_item(self, row: list, cls: list) -> str:
        assert len(row) == len(cls)
        return f"| {' | '.join([self._fmt(*r) for r in zip(row, cls)])} |\n"

    def type_end(self) -> str:
        return ''

    # Utility functions
    def _fmt(self, s: str, f: str) -> str:
        f1 = {'n': '', 's': '', 'd': '', 'b': '**', 'h': '**_'}
        f2 = {'n': '', 's': '', 'd': '', 'b': '**', 'h': '_**'}
        ss = '\\*' if s == '*' else s
        return f'{f1[f]}{ss}{f2[f]}'


# ---------- JADN Source (JAS) output ------------------
class JasConversion(BasicConversion):
    def doc_begin(self) -> str:
        return ''

    def doc_end(self) -> str:
        return ''

    def sect(self, num, name) -> str:
        return ''

    def meta_begin(self) -> str:
        return ''

    def meta_item(self, key: str, val: Union[dict, list, str]) -> str:
        return ''

    def meta_end(self) -> str:
        return ''

    def type_begin(self, tname: str, ttype: str, headers: list, cls: list) -> str:
        assert len(headers) == len(cls)
        return ''

    def type_item(self, row: list, cls: list) -> str:
        assert len(row) == len(cls)
        return ''

    def type_end(self) -> str:
        return ''


# ---------- CDDL output ------------------
class CddlConversion(BasicConversion):
    def doc_begin(self) -> str:
        return ''

    def doc_end(self) -> str:
        return ''

    def sect(self, num, name) -> str:
        return ''

    def meta_begin(self) -> str:
        return ''

    def meta_item(self, key: str, val: Union[dict, list, str]) -> str:
        return ''

    def meta_end(self) -> str:
        return ''

    def type_begin(self, tname: str, ttype: str, headers: list, cls: list) -> str:
        assert len(headers) == len(cls)
        return ''

    def type_item(self, row: list, cls: list) -> str:
        assert len(row) == len(cls)
        return ''

    def type_end(self) -> str:
        return ''


# ---------- Thrift output ------------------
class ThriftConversion(BasicConversion):
    def doc_begin(self) -> str:
        return ''

    def doc_end(self, ) -> str:
        return ''

    def sect(self, num, name) -> str:
        return ''

    def meta_begin(self) -> str:
        return ''

    def meta_item(self, key: str, val: Union[dict, list, str]) -> str:
        return ''

    def meta_end(self, ) -> str:
        return ''

    def type_begin(self, tname: str, ttype: str, headers: list, cls: list) -> str:
        assert len(headers) == len(cls)
        return ''

    def type_item(self, row: list, cls: list) -> str:
        assert len(row) == len(cls)
        return ''

    def type_end(self) -> str:
        return ''


# ----------------------------------------------
class ConversionFormats(str, Enum):
    CDDL = 'cddl'
    JAS = 'jas'
    MarkDown = 'markdown'
    Thrift = 'thrift'


wtab: Dict[ConversionFormats, Type[BasicConversion]] = {
    ConversionFormats.CDDL: CddlConversion,
    ConversionFormats.JAS: JasConversion,
    ConversionFormats.MarkDown: MarkdownConversion,
    ConversionFormats.Thrift: ThriftConversion
}

DEFAULT_FORMAT = ConversionFormats.MarkDown


def table_dumps(schema: dict, form: ConversionFormats = DEFAULT_FORMAT) -> str:
    if cls := wtab.get(form):
        return cls(schema).convert()
    raise ValueError(f'{form} is not a valid conversion format')


def table_dump(schema: dict, fname: Union[bytes, str, int], source='', form: ConversionFormats = DEFAULT_FORMAT) -> NoReturn:
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write(f'<!-- Generated from {source}, {datetime.ctime(datetime.now())}-->\n')
        f.write(table_dumps(schema, form))


__all__ = [
    'ConversionFormats',
    'table_dump',
    'table_dumps'
]



"""
Support functions for JADN codec
  Convert dict between nested and flat
  Convert typedef options between dict and strings
"""
import copy
import re

from functools import reduce
from typing import Any, Dict, List, NoReturn, Set, Tuple, Union
from .definitions import (
    TypeName, BaseType, TypeOptions, Fields, ItemDesc, FieldID, FieldName, FieldType, FieldOptions, FieldDesc,
    DEFAULT_CONFIG, TYPE_OPTIONS, FIELD_OPTIONS, OPTION_ID, OPTION_TYPES, is_builtin, has_fields, TypeDefinition,
    EnumFieldDefinition, GenFieldDefinition
)


# Handle errors
def raise_error(*s) -> NoReturn:
    raise ValueError(*s)


# Truncate a string to "n" characters, replacing end with ".." if truncated
def etrunc(s: str, n: int) -> str:
    return s if n is None else s[:n-2] + (s[n-2:], '..')[len(s) > n] if n > 1 else s[:n]


# Dict conversion utilities
def dmerge(*dicts: dict) -> dict:
    """
    Merge any number of dicts
    """
    return {k: v for d in dicts for k, v in d.items()}


def hdict(keys: str, value: any, sep: str = '.') -> dict:
    """
    Convert a hierarchical-key value pair to a nested dict
    """
    return reduce(lambda v, k: {k: v}, reversed(keys.split(sep)), value)


def fluff(src: dict, sep: str = '.') -> dict:
    """
    Convert a flat dict with hierarchical keys to a nested dict

    :param src: flat dict - e.g., {'a.b.c': 1, 'a.b.d': 2}
    :param sep: separator character for keys
    :return: nested dict - e.g., {'a': {'b': {'c': 1, 'd': 2}}}
    """
    return reduce(dmerge, [hdict(k, v, sep) for k, v in src.items()], {})


def flatten(cmd: dict, path: str = '', fc: dict = None, sep: str = '.') -> dict:
    """
    Convert a nested dict to a flat dict with hierarchical keys
    """
    if fc is None:
        fc = {}
    fcmd = fc.copy()
    if isinstance(cmd, dict):
        for k, v in cmd.items():
            k = k.split(':')[1] if ':' in k else k
            fcmd = flatten(v, sep.join((path, k)) if path else k, fcmd)
    elif isinstance(cmd, list):
        for n, v in enumerate(cmd):
            fcmd.update(flatten(v, sep.join([path, str(n)])))
    else:
        fcmd[path] = cmd
    return fcmd


def dlist(src: dict) -> dict:
    """
    Convert dicts with numeric keys to lists

    :param src: {'a': {'b': {'0':'red', '1':'blue'}, 'c': 'foo'}}
    :return: {'a': {'b': ['red', 'blue'], 'c': 'foo'}}
    """
    if isinstance(src, dict):
        for k in src:
            src[k] = dlist(src[k])
        if set(src) == {str(k) for k in range(len(src))}:
            src = [src[str(k)] for k in range(len(src))]
    return src


def build_deps(schema: dict) -> Dict[str, Set[str]]:
    """
    Build a Dependency dict: {TypeName: {Dep1, Dep2, ...}}
    """
    # Options whose value is/has a type name
    oids = [OPTION_ID['ktype'], OPTION_ID['vtype'], OPTION_ID['and'], OPTION_ID['or']]
    # Options that enumerate fields
    oids2 = [OPTION_ID['enum'], OPTION_ID['pointer']]

    def get_refs(tdef: list) -> Set[str]:  # Return all type references from a type definition
        refs = {to[1:] for to in tdef[TypeOptions] if to[0] in oids and not is_builtin(to[1:])}
        refs.update([to for to in tdef[TypeOptions] if to[0] in oids2])
        if has_fields(tdef[BaseType]):
            for f in tdef[Fields]:
                if not is_builtin(f[FieldType]):
                    # Add reference to type name
                    refs.add(f[FieldType])
                # Get refs from type opts in field (extension)
                refs.update(get_refs(['', f[FieldType], f[FieldOptions], '']))
        return refs

    return {t[TypeName]: get_refs(t) for t in schema['types']}


def topo_sort(items: List[Tuple[str, List[str]]]) -> Tuple[List[str], Set[str]]:
    """
    Topological sort with locality
    Sorts a list of (item: (dependencies)) pairs so that 1) all dependency items are listed before the parent item,
    and 2) dependencies are listed in the given order and as close to the parent as possible.
    Returns the sorted list of items and a list of root items.  A single root indicates a fully-connected hierarchy;
    multiple roots indicate disconnected items or hierarchies, and no roots indicate a dependency cycle.
    """
    out: List[str] = []
    deps = {i[0]: i[1] for i in items}  # TODO: update for items dict

    def walk_tree(it: str) -> NoReturn:
        for i in deps[it]:
            if i not in out:
                walk_tree(i)
                out.append(i)

    roots = {i[0] for i in items} - {i[1] for i in items}
    for item in roots:
        walk_tree(item)
        out.append(item)
    out = out if out else [i[0] for i in items]     # if cycle detected, don't sort
    return out, roots


def get_optx(opts: List[OPTION_TYPES], oname: str) -> Union[OPTION_TYPES, None]:
    if n := [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]:
        return n[0]
    return None


def del_opt(opts: List[OPTION_TYPES], oname: str) -> NoReturn:
    if n := [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]:
        del opts[n[0]]


def topts_s2d(olist: Union[List[OPTION_TYPES], Tuple[OPTION_TYPES, ...]], frange: bool = False) -> dict:
    """
    Convert list of type definition option strings to options dictionary
    """
    assert isinstance(olist, (list, tuple)), f'{olist} is not a list'
    topts = {o for o in olist if ord(o[0]) in TYPE_OPTIONS}
    if uopts := {*olist} - topts:
        raise_error(f"Unknown type options: {','.join(uopts)}")
    opts = {}
    for o in topts:
        k, v, _ = TYPE_OPTIONS[ord(o[0])]
        opts[k] = v(o[1:])
    return opts


def ftopts_s2d(olist: Union[List[OPTION_TYPES], Tuple[OPTION_TYPES, ...]]) -> Tuple[dict, dict]:
    """
    Convert list of field definition option strings to options dictionary
    returns - FieldOptions, TypeOptions
    """
    assert isinstance(olist, (list, tuple)), f'{olist} is not a list'
    fopts = {}
    topts = {}
    for o in olist:
        try:
            k, v, _ = FIELD_OPTIONS[ord(o[0])]
            fopts[k] = v(o[1:])
        except KeyError:
            topts.update(topts_s2d([o]))
    return fopts, topts


def opts_d2s(to: dict) -> List[str]:
    rtn: List[str] = []
    for k, v in to.items():
        try:
            rtn.append(f"{OPTION_ID[k]}{'' if v is True else str(v)}")
        except KeyError:
            raise_error(f'Unknown option tag {k}')
    return rtn


def opts_sort(olist: Union[List[OPTION_TYPES], Tuple[OPTION_TYPES, ...]]) -> NoReturn:
    """
    Sort JADN option list into canonical order
    """
    def opt_order(o):
        try:
            k = FIELD_OPTIONS[ord(o)][2]
        except KeyError:
            k = TYPE_OPTIONS[ord(o)][2]
        return k

    olist.sort(key=lambda x: opt_order(x[0]))


def canonicalize(schema: dict) -> dict:
    def can_opts(olist: List[OPTION_TYPES], btype: str):
        opts_sort(olist)                # Sort options into canonical order (for comparisons)
        fo, to = ftopts_s2d(olist)      # Remove default size and multiplicity options
        if 'minv' in to and to['minv'] == 0 and btype != 'Integer':
            del_opt(olist, 'minv')
        if 'minc' in fo and fo['minc'] == 1:
            del_opt(olist, 'minc')
        if 'maxc' in fo and fo['maxc'] == 1:
            del_opt(olist, 'maxc')
        if btype == 'Number':           # TODO: fix corner case input = 2.000
            minf = get_optx(olist, 'minf')
            if minf is not None and '.' not in olist[minf]:
                olist[minf] += '.0'
            maxf = get_optx(olist, 'maxf')
            if maxf is not None and '.' not in olist[maxf]:
                olist[maxf] += '.0'

    cschema = copy.deepcopy(schema)     # don't modify original
    for td in cschema['types']:
        can_opts(td[TypeOptions], td[BaseType])
        for fd in td[Fields]:
            if td[BaseType] != 'Enumerated':
                can_opts(fd[FieldOptions], fd[FieldType])
    return cschema


def cleanup_tagid(fields: List[list]) -> List[list]:
    """
    If type definition contains a TagId option, replace field name with id
    """
    for f in fields:
        if len(f) > FieldOptions:
            tx = get_optx(f[FieldOptions], 'tagid')
            if tx is not None:
                to = f[FieldOptions][tx]
                try:
                    int(to[1:])                 # Check if already a Field Id
                except ValueError:              # Look up Id corresponding to Field Name
                    fx = {x[FieldName]: x[FieldID] for x in fields}
                    f[FieldOptions][tx] = to[0] + str(fx[to[1:]])
    return fields


def typestr2jadn(typestring: str) -> Tuple[str, List[str], list]:
    def parseopt(optstr: str) -> str:
        m1 = re.match(r'^\s*([-$\w]+)(?:\[([^]]+)\])?$', optstr)
        if m1 is None:
            raise_error(f'TypeString2JADN: unexpected function: {optstr}')
        return OPTION_ID[m1.group(1).lower()] + m1.group(2) if m1.group(2) else m1.group(1)

    topts = {}
    fo = []
    p_name = r'\s*=?\s*([-$:\w]+)'      # 1 type name
    p_id = r'(\.ID)?'                   # 2 'id'
    p_func = r'(?:\(([^)]+)\))?'        # 3 'ktype', 'vtype', 'enum', 'pointer', 'tagid'
    p_rangepat = r'(?:\{(.*)\})?'       # 4 'minv', 'maxv', 'pattern'
    p_format = r'(?:\s+\/(\w[-\w]*))?'  # 5 'format'
    p_unique = r'(\s+unique)?'          # 6 'unique'
    pattern = fr'^{p_name}{p_id}{p_func}{p_rangepat}{p_format}{p_unique}\s*$'
    m = re.match(pattern, typestring)
    if m is None:
        raise_error(f'TypeString2JADN: "{typestring}" does not match pattern {pattern}')
    tname = m.group(1)
    topts.update({'id': True} if m.group(2) else {})
    if m.group(3):                      # (ktype, vtype), Enum(), Pointer() options
        opts = [parseopt(x) for x in m.group(3).split(',', maxsplit=1)]
        if tname == 'MapOf':
            topts.update({'ktype': opts[0], 'vtype': opts[1]})
        elif tname == 'ArrayOf':
            assert len(opts) == 1
            topts.update({'vtype': opts[0]})
        else:
            assert len(opts) == 1
            topts.update(topts_s2d([opts[0]]) if ord(opts[0][0]) in TYPE_OPTIONS else {})
            fo += [opts[0]] if ord(opts[0][0]) in FIELD_OPTIONS else []         # TagId option
    if m.group(4):
        if m1 := re.match('^pattern="(.+)"$', m.group(4)):
            topts.update({'pattern': m1.group(1)})
        else:
            a, b = m.group(4).split('..', maxsplit=1)
            if tname == 'Number':
                topts.update({} if a == '*' else {'minf': float(a)})
                topts.update({} if b == '*' else {'maxf': float(b)})
            else:
                a = '*' if tname != 'Integer' and a != '*' and int(a) == 0 else a   # Default min size = 0
                topts.update({} if a == '*' else {'minv': int(a)})
                topts.update({} if b == '*' else {'maxv': int(b)})
    topts.update({'format': m.group(5)} if m.group(5) else {})
    topts.update({'unique': True} if m.group(6) else {})
    return tname, opts_d2s(topts), fo


def jadn2typestr(tname: str, topts: List[OPTION_TYPES]) -> str:
    """
    Convert typename and options to string
    """
    # Handle ktype/vtype containing Enum options
    def _kvstr(optv: str) -> str:
        if optv[0] == OPTION_ID['enum']:
            return f'Enum[{optv[1:]}]'
        if optv[0] == OPTION_ID['pointer']:
            return f'Pointer[{optv[1:]}]'
        return optv

    # Size range (single-ended) - default is {0..*}
    def _srange(ops: dict) -> str:
        lo = ops.pop('minv', 0)
        hi = ops.pop('maxv', -1)
        hs = '*' if hi < 0 else str(hi)
        return f'{lo}..{hs}' if lo != 0 or hs != '*' else ''

    # Value range (double-ended) - default is {*..*}
    def _vrange(ops: dict) -> str:
        lo = ops.pop('minv', '*')
        hi = ops.pop('maxv', '*')
        return f'{lo}..{hi}' if lo != '*' or hi != '*' else ''

    # Value range (double-ended) - default is {*..*}
    def _frange(ops: dict) -> str:
        lo = ops.pop('minf', '*')
        hi = ops.pop('maxf', '*')
        return f'{lo}..{hi}' if lo != '*' or hi != '*' else ''

    opts = topts_s2d(topts)
    extra = '.ID' if opts.pop('id', None) else ''   # SIDE EFFECT: remove known options from opts.
    if tname in ('ArrayOf', 'MapOf'):
        extra += f"({_kvstr(opts.pop('ktype'))}, " if tname == 'MapOf' else '('
        extra += f"{_kvstr(opts.pop('vtype'))})"

    if v := opts.pop('enum', None):
        extra += f'(Enum[{v}])'

    if v := opts.pop('pointer', None):
        extra += f'(Pointer[{v}])'

    if v := opts.pop('pattern', None):  # String can have {range} or {pattern} or /format
        extra += f'{{pattern="{v}"}}'

    if v := _vrange(opts) if tname == 'Integer' else (_frange(opts) if tname == 'Number' else _srange(opts)):
        extra += f'{{{v}}}'

    if v := opts.pop('format', None):
        extra += f' /{v}'

    if opts.pop('unique', None):
        extra += ' unique'

    if v := opts.pop('and', None):  # hack set operations for now.  TODO: generalize to any number
        extra += f' ∩ {v}'

    if v := opts.pop('or', None):
        extra += f' ∪ {v}'

    return f"{tname}{extra}{f' ?{str(map(str, opts))}?' if opts else ''}"  # Flag unrecognized options


def jadn2fielddef(fdef: list, tdef: list) -> Tuple[str, str, str, str]:
    idtype = tdef[BaseType] == 'Array' or get_optx(tdef[TypeOptions], 'id') is not None
    fname = '' if idtype else fdef[FieldName]
    fdesc = f'{fdef[FieldName]}:: ' if idtype else ''
    is_enum = tdef[BaseType] == 'Enumerated'
    fdesc += fdef[ItemDesc if is_enum else FieldDesc]
    ftyperef = ''
    fmult = ''

    if not is_enum:
        fo, fto = ftopts_s2d(fdef[FieldOptions])
        fname += '/' if 'dir' in fo else ''
        tf = ''
        if tagid := fo.get('tagid', None):
            tf = [f[FieldName] for f in tdef[Fields] if f[FieldID] == tagid][0]
            tf = f'(TagId[{tf if tf else tagid}])'
        ft = jadn2typestr(f'{fdef[FieldType]}{tf}', opts_d2s(fto))
        ftyperef = f'Key({ft})' if 'key' in fo else f'Link({ft})' if 'link' in fo else ft
        minc = fo.get('minc', 1)
        maxc = fo.get('maxc', 1)
        fmult = '1' if minc == 1 and maxc == 1 else f"{minc}..{'*' if maxc == 0 else maxc}"
    return fname, ftyperef, fmult, fdesc


def fielddef2jadn(fid: int, fname: str, fstr: str, fmult: str, fdesc: str) -> list:
    ftyperef = ''
    fo = {}
    if fstr:
        if m := re.match(r'^(Link|Key)\((.*)\)$', fstr):
            fo = {m.group(1).lower(): True}
            fstr = m.group(2)
        ftyperef, topts, fopts = typestr2jadn(fstr)
        # Field is one of: enum.id, enum, field.id, field
        fo.update(topts_s2d(topts))                   # Copy type options (if any) into field options (JADN extension)
        if fname.endswith('/'):
            fo.update({'dir': True})
            fname = fname.rstrip('/')
        if m := re.match(r'^(\d+)(?:\.\.(\d+|\*))?$', fmult) if fmult else None:
            groups = m.groups()
            if maxc := groups[1]:
                minc = int(groups[0])
                maxc = 0 if maxc == '*' else int(maxc)
            else:
                minc = maxc = int(groups[0])
            fo.update({'minc': minc} if minc != 1 else {})
            fo.update({'maxc': maxc} if maxc != 1 else {})
        elif fmult:
            fo.update({'minc': -1, 'maxc': -1})
        if fopts:
            assert len(fopts) == 1 and fopts[0][0] == OPTION_ID['tagid']    # Update if additional field options defined
            fo.update({'tagid': fopts[0][1:]})      # if field name, MUST update to id after all fields have been read
    if fdesc:
        m = re.match(r'^(?:\s*\/\/)?\s*(.*)$', fdesc)
        fdesc = m.group(1)
        if not fname:
            if m := re.match(r'^([^:]+)::\s*(.*)$', fdesc):
                fname = m.group(1)
                fdesc = m.group(2)
    return [fid, fname, ftyperef, opts_d2s(fo), fdesc] if ftyperef else [fid, fname, fdesc]


def get_config(meta: dict) -> dict:
    config = dict(DEFAULT_CONFIG)
    if meta:
        config.update(meta.get('config', {}))
    return config


# Schema conversion for object-like use
def object_types(types: List[list]) -> List[TypeDefinition]:
    rtn_types: List[TypeDefinition] = []
    for t in types:
        t = TypeDefinition(*t)
        if t.BaseType == 'Enumerated':
            t.Fields = [EnumFieldDefinition(*f) for f in t.Fields]
        else:
            t.Fields = [GenFieldDefinition(*f) for f in t.Fields]
        rtn_types.append(t)
    return rtn_types


def object_type_schema(schema: dict) -> dict:
    sc = copy.deepcopy(schema)
    sc['types'] = object_types(sc['types'])
    return sc


def list_types(types: List[TypeDefinition]) -> List[list]:
    return [[*t[:-1], [list(f) for f in t.Fields]] for t in types]


def list_type_schema(schema: dict) -> dict:
    sc = copy.deepcopy(schema)
    sc['types'] = list_types(sc['types'])
    return sc


# General Utilities
def list_get_default(lst: list, idx: int, default: Any = None) -> Any:
    try:
        return lst[idx]
    except IndexError:
        return default



from .html import html_dump, html_dumps, html_load, html_loads
from .jidl import jidl_dump, jidl_dumps, jidl_load, jidl_loads, jidl_style
from .table_w import table_dump, table_dumps, ConversionFormats
from .dot_w import dot_dump, dot_dumps, dot_style
from .jas_w import jas_dump, jas_dumps



"""
Translate JADN to JAS (JADN Abstract Syntax)
"""
from copy import deepcopy
from datetime import datetime
from textwrap import fill
from typing import NoReturn, Union
from ..definitions import (
    TypeName, BaseType, TypeOptions, TypeDesc, Fields, ItemDesc, FieldID, FieldName, FieldType, FieldOptions, FieldDesc,
    CORE_TYPES, INFO_ORDER, TYPE_OPTIONS, FIELD_OPTIONS
)
from ..utils import ftopts_s2d, topts_s2d

stype_map = {                   # Map JADN built-in types to JAS type names (Equivalent ASN.1 types in comments)
    'Binary': 'BINARY',         # OCTET STRING
    'Boolean': 'BOOLEAN',       # BOOLEAN
    'Integer': 'INTEGER',       # INTEGER
    'Number': 'REAL',           # REAL
    'Null': 'NULL',             # NULL
    'String': 'STRING',         # UTF8String
    'Array': 'ARRAY',           # SEQUENCE
    'ArrayOf': 'ARRAY_OF',      # SEQUENCE OF
    'Choice': 'CHOICE',         # CHOICE
    'Enumerated': 'ENUMERATED', # ENUMERATED
    'Map': 'MAP',               # SET
    'MapOf': 'MAP_OF',          #
    'Record': 'RECORD'          # SEQUENCE
}


def stype(jtype: str) -> str:
    return stype_map.get(jtype, jtype)


# TODO: simplify function??
def jas_dumps(schema: dict) -> str:
    """
    Produce JAS module from JADN structure

    JAS represents features available in both JADN and ASN.1 using ASN.1 syntax, but adds
    extended datatypes (Record, Map) for JADN types not directly representable in ASN.1.
    With appropriate encoding rules (which do not yet exist), SEQUENCE could replace Record.
    Map could be implemented using ASN.1 table constraints, but for the purpose of representing
    JSON objects, the Map first-class type in JAS is easier to use.
    """
    jas = ''

    # Convert Meta
    if info := schema.get('info'):
        jas += '/*\n'
        mlist = [k for k in INFO_ORDER if k in info]
        for h in mlist + list(set(info) - set(mlist)):
            if h == 'description':
                jas += fill(info[h], width=80, initial_indent='{0:14} '.format(h+':'), subsequent_indent=15*' ') + '\n'
            elif h == 'imports':
                hh = '{:14} '.format(f'{h}:')
                for imp in info[h]:
                    jas += '{}{}: {}\n'.format(hh, *imp)
                    hh = 15*' '
            elif h == 'exports':
                jas += '{:14} {}\n'.format(f'{h}:', ', '.join(info[h]))
            else:
                jas += '{:14} {}\n'.format(f'{h}:', info[h])
        jas += '*/\n'

    assert set(stype_map) == set(CORE_TYPES)         # Ensure type list is up to date
    tolist = {'id', 'vtype', 'ktype', 'enum', 'pointer', 'format', 'pattern', 'minv', 'maxv', 'minf', 'maxf', 'unique', 'and', 'or'}
    assert {x[0] for x in TYPE_OPTIONS.values()} == tolist                # Ensure type options list is up to date
    folist = {'minc', 'maxc', 'tagid', 'dir', 'key', 'link', 'default'}
    assert {x[0] for x in FIELD_OPTIONS.values()} == folist               # Ensure field options list is up to date

    # Convert Types
    for td in schema['types']:  # 0:type name, 1:base type, 2:type opts, 3:type desc, 4:fields
        topts = topts_s2d(td[TypeOptions])
        tostr = ''
        v_range = ''
        if 'minv' in topts or 'maxv' in topts:          # TODO: use jadn2typestr
            lo = topts.get('minv', 0)
            hi = topts.get('maxv', 0)
            if lo or hi:
                v_range = f"({lo}..{hi if hi else 'MAX'})"
        for opt in tolist:
            if opt in topts:
                ov = topts[opt]
                if opt == 'id':
                    tostr += '.ID'
                elif opt == 'vtype':
                    tostr += f'({ov})'
                elif opt == 'ktype':
                    pass            # fix MapOf(ktype, vtype)
                elif opt == 'pattern':
                    tostr += f' (PATTERN ("{ov}"))'
                elif opt == 'format':
                    tostr += f' (CONSTRAINED BY {{{ov}}})'
                elif opt in ('minv', 'maxv'):     # TODO fix to handle both
                    if v_range:
                        if td[BaseType] in ('Integer', 'Number'):
                            tostr += f' {v_range}'
                        elif td[BaseType] in ('Binary', 'String', 'Array', 'ArrayOf', 'Map', 'MapOf', 'Record'):
                            tostr += f' (Size {v_range})'
                        else:
                            assert False        # Should never get here
                    v_range = ''
                else:
                    tostr += f' %{opt}: {ov}%'
        tdesc = f'    -- {td[TypeDesc]}' if td[TypeDesc] else ''
        jas += f'\n{td[TypeName]} ::= {stype(td[BaseType])}{tostr}'
        if len(td) > Fields:
            titems = deepcopy(td[Fields])
            for n, i in enumerate(titems):      # 0:tag, 1:enum item name, 2:enum item desc  (enumerated), or
                if len(i) > FieldOptions:              # 0:tag, 1:field name, 2:field type, 3: field opts, 4:field desc
                    desc = i[FieldDesc]
                    i[FieldType] = stype(i[FieldType])
                else:
                    desc = i[ItemDesc]
                desc = f'    -- {desc}' if desc else ''
                i.append(',' + desc if n < len(titems) - 1 else (' ' + desc if desc else ''))  # TODO: fix hacked desc for join
            flen = min(32, max(12, max([len(i[FieldName]) for i in titems]) + 1 if titems else 0))
            jas += ' {' + tdesc + '\n'
            if td[BaseType].lower() == 'enumerated':
                fmt = '    {1:' + str(flen) + '} ({0:d}){3}'
                jas += '\n'.join([fmt.format(*i) for i in titems])
            else:
                fmt = '    {1:' + str(flen) + '} [{0:d}] {2}{3}{4}'
                if td[BaseType].lower() == 'record':
                    fmt = '    {1:' + str(flen) + '} {2}{3}{4}'
                items = []
                for n, i in enumerate(titems):                          # TODO: Convert to use jadn2fielddef
                    ostr = ''
                    opts = ftopts_s2d(i[FieldOptions])[0]
                    if 'tagid' in opts:
                        ostr += f"(Tag({opts['tagid']}))"    # TODO: lookup field name
                        del opts['tagid']
                    if 'vtype' in opts:
                        ostr += '.*'
                        del opts['vtype']
                    if 'minc' in opts:
                        if opts['minc'] == 0:         # TODO: handle array fields (max != 1)
                            ostr += ' OPTIONAL'
                        del opts['minc']
                    items += [fmt.format(i[FieldID], i[FieldName], i[FieldType], ostr, i[5]) + (f' %{opts}' if opts else '')]
                jas += '\n'.join(items)
            jas += '\n}\n' if titems else '}\n'
        else:
            jas += tdesc + '\n'
    return jas


def jas_dump(schema: dict, fname: Union[bytes, str, int], source='') -> NoReturn:
    with open(fname, 'w') as f:
        if source:
            f.write(f'-- Generated from {source}, {datetime.ctime(datetime.now())}\n\n')
        f.write(jas_dumps(schema))


__all__ = [
    'jas_dump',
    'jas_dumps'
]



import re
import json

from datetime import datetime
from typing import Generator, NoReturn, Tuple, Union
from lxml import html
from .utils import DocHTML
from ..definitions import Fields, ItemID, ItemDesc, FieldID, INFO_ORDER, TypeDefinition
from ..utils import cleanup_tagid, fielddef2jadn, jadn2fielddef, jadn2typestr, typestr2jadn
"""
Convert JADN schema to and from HTML format

Generated HTML uses "div tables" rather than HTML table elements to support responsive page design.
Css styles for classes beginning with 't' must be defined to render correctly.

Generated HTML includes JADN Schema classes to support parsing.
Css styles for classes beginning with 'j' are not required

CSS class definitions needed to render table layout:
tTable:   display table
tCaption: display table-caption
tHead:    display table-header-group
tHCell:   display table-cell (used in header)
tBody:    display table-row-group
tRow:     display table-row
tCell:    display table-cell (used in body)

JADN schema elements (required for parsing, available for styling):
jKey:     metadata key
jVal:     metadata value
jTname:   TypeName - name of a defined type
jTstr:    String representation of BaseType and TypeOptions
jTdesc:   TypeDesc - description of a type
jFid:     FieldId - integer tag of a field or enumerated item
jFname:   FieldName/ItemValue - name of a field or string value of an enumerated item
jFstr:    String representation of FieldType and FieldOptions (except multiplicity)
jFmult:   String representation of multiplicity FieldOptions
jFdesc:   FieldDesc/ItemDesc - description of a field in a container type or item in an enumerated type

JADN styling classes (not used for parsing):
jMeta:    metadata section
jType:    type definition
jHid:     FieldId
jHname:   FieldName
jHstr:    FieldType + FieldOptions
jHmult:   Multiplicity
jHdesc:   Description
"""


# Convert JADN to HTML
def html_dumps(schema: dict) -> str:
    # Make initial tree
    doc, tag = DocHTML('<!DOCTYPE html>', lang='en').context()

    with tag('head'):
        tag('meta', charset='UTF-8')
        tag('title', schema.get('info', {}).get('title', 'JADN Schema'))
        tag('link', rel='stylesheet', href='css/dtheme.css', type='text/css')

    with tag('body'):
        tag('h2', 'Schema')

        # Add meta elements if present
        if info := schema.get('info', None):
            with tag('div', klass='tBody'):
                mlist = [k for k in INFO_ORDER if k in info]
                for k in mlist + list({*info} - {*mlist}):
                    with tag('div', klass='tRow'):
                        tag('div', f'{k}:', klass='tCell jKey')
                        tag('div', json.dumps(info[k]), klass='tCell jVal')
                # top-level element of the metadata table
                tag('div', klass='tTable jinfo')

        # Add type definitions
        for tdef in schema['types']:
            tdef = TypeDefinition(*tdef)
            with tag('div', klass='tTable jType'):  # top-level element of a type definition
                with tag('div', klass='tCaption'):
                    with tag('div', klass='jTdef'):  # container for type definition column
                        tag('div', tdef.TypeName, klass='jTname')
                        tag('div', f' = {jadn2typestr(tdef.BaseType, tdef.TypeOptions)}', klass='jTstr')
                    tag('div', tdef.TypeDesc or '', klass='jTdesc')

                if len(tdef) > Fields:
                    with tag('div', klass='tHead'):
                        tag('div', 'ID', klass='tHCell jHid')
                        tag('div', 'Name', klass='tHCell jHname')
                        if tdef.Fields and len(tdef.Fields[0]) > ItemDesc + 1:
                            tag('div', 'Type', klass='tHCell jHstr')
                            tag('div', '#', klass='tHCell jHmult')
                        tag('div', 'Description', klass='tHCell jHdesc')

                    with tag('div', klass='tBody'):
                        for fdef in tdef.Fields:
                            with tag('div', klass='tRow'):
                                fname, ftyperef, fmult, fdesc = jadn2fielddef(fdef, [*tdef])
                                if len(tdef.Fields[0]) > ItemDesc + 1:
                                    tag('div', str(fdef[FieldID]), klass='tCell jFid')
                                    tag('div', fname, klass='tCell jFname')
                                    tag('div', ftyperef, klass='tCell jFstr')
                                    tag('div', fmult, klass='tCell jFmult')
                                else:
                                    tag('div', str(fdef[ItemID]), klass='tCell jFid')
                                    tag('div', fname, klass='tCell jFname')
                                tag('div', fdesc, klass='tCell jFdesc')
    return doc.getvalue(True)


def html_dump(schema: dict, fname: Union[bytes, str, int], source='') -> NoReturn:
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write(f'<! Generated from {source}, {datetime.ctime(datetime.now())}>\n\n')
        f.write(html_dumps(schema))


# Convert HTML to JADN
def line2jadn(lt: dict, tdef) -> Tuple[str, list]:
    def default(x: str, d: str) -> str:
        return x if x else d

    if 'jKey' in lt:
        return 'M', [lt['jKey'].rstrip(':'), lt['jVal']]
    if 'jTname' in lt:
        btype, topts, fo = typestr2jadn(lt['jTstr'])
        assert fo == []                     # field options MUST not be included in typedefs
        tdesc = default(lt.get('jTdesc', ''), '')
        if tdesc:
            m = re.match(r'^(?:\s*\/\/)?\s*(.*)$', tdesc)
            tdesc = m.group(1)
        return 'T', [lt['jTname'], btype, topts, tdesc, []]
    if 'jFid' in lt:
        fname = default(lt.get('jFname', ''), '')
        fstr = default(lt.get('jFstr', ''), '')
        fdesc = default(lt.get('jFdesc', ''), '')
        fmult = lt.get('jFmult', '')
        return 'F', fielddef2jadn(int(lt['jFid']), fname, fstr, fmult, fdesc)


def html_loads(hdoc: str) -> dict:
    tclass = ('jKey', 'jVal', 'jTname', 'jTstr', 'jTdesc', 'jFid', 'jFname', 'jFstr', 'jFmult', 'jFdesc')

    # TODO: convert to use minidom
    def get_line(hdoc: str) -> Generator[dict, None, None]:
        line = {}
        for element in html.fromstring(hdoc).iter():
            cl = [c for c in element.get('class', '').split() if c in tclass]  # Get element's token class
            if cl:
                assert len(cl) == 1  # TODO: Can't be more than one - replace assertions with proper errors
                if cl[0] in ('jKey', 'jTname', 'jFid'):
                    if line:
                        yield line
                        line = {}
                line.update({cl[0]: element.text})
        yield line

    meta = {}
    types = []
    fields = None
    for line in get_line(hdoc):
        if line:
            t, v = line2jadn(line, types[-1] if types else None)
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                meta.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'info': meta, 'types': types} if meta else {'types': types}


def html_load(fname: Union[bytes, str, int]) -> dict:
    with open(fname, 'r') as f:
        return html_loads(f.read())


__all__ = [
    'html_dump',
    'html_dumps',
    'html_load',
    'html_loads'
]



import copy

from typing import Generator, List, NoReturn, Set, Union
from ..definitions import (
    TypeName, BaseType, TypeDesc, Fields, ItemID, ItemValue, ItemDesc, FieldName, FieldOptions, FieldDesc, OPTION_ID,
    EXTENSIONS, OPTION_TYPES, is_builtin, has_fields, TypeDefinition, EnumFieldDefinition, GenFieldDefinition
)
from ..utils import del_opt, ftopts_s2d, get_optx, list_type_schema, opts_d2s, object_type_schema, topts_s2d, etrunc


def strip_comments(schema: dict, width=0) -> dict:  # Strip or truncate comments from schema
    sc = copy.deepcopy(schema)
    for tdef in sc['types']:
        tdef[TypeDesc] = etrunc(tdef[TypeDesc], width)
        if len(tdef) > Fields:
            fd = ItemDesc if tdef[BaseType] == 'Enumerated' else FieldDesc
            for fdef in tdef[Fields]:
                fdef[fd] = etrunc(fdef[fd], width)
    return sc


# Simplify schema
# Return option array index of enum or pointer option
def epx(topts: List[OPTION_TYPES]) -> Union[int, float, str, None]:
    ex = get_optx(topts, 'enum')
    return ex if ex is not None else get_optx(topts, 'pointer')


def epname(topts: List[OPTION_TYPES], sys: str = '$') -> Union[str, None]:
    x = epx(topts)
    if x is None:
        return None
    oname = 'Enum' if topts[x][0] == OPTION_ID['enum'] else 'Pointer'
    return f"{topts[x][1:]}{sys}{oname}{'-Id' if get_optx(topts, 'id') else ''}"


# Replace field multiplicity with explicit ArrayOf type definitions
def simplify_multiplicity(schema: dict, sys: str = '$') -> NoReturn:
    for tdef in list(schema['types']):
        if has_fields(tdef.BaseType):
            for fdef in tdef.Fields:
                fo, fto = ftopts_s2d(fdef.FieldOptions)
                if 'maxc' in fo and fo['maxc'] != 1:
                    minc = fo.get('minc', 1)
                    newopts = {
                        'vtype': fdef.FieldType,
                        'minv': max(minc, 1),  # Don't allow empty ArrayOf
                        **({'maxv': fo['maxc']} if fo['maxc'] > 1 else {}),  # maxv defaults to 0
                        **({'unique': True} if 'unique' in fto else {})  # Move unique option to ArrayOf
                    }
                    # Point existing field to new ArrayOf
                    fdef.FieldType = f'{tdef.TypeName}{sys}{fdef.FieldName}'
                    schema['types'].append(TypeDefinition(fdef.FieldType, 'ArrayOf', opts_d2s(newopts), fdef.FieldDesc))
                    # Remove unused FieldOptions
                    del_opt(fdef.FieldOptions, 'maxc')
                    if minc != 0:
                        del_opt(fdef.FieldOptions, 'minc')
                    del_opt(fdef.FieldOptions, 'unique')


# Replace anonymous types in fields with explicit type definitions
def simplify_anonymous_types(schema: dict, sys: str = '$') -> NoReturn:
    for tdef in list(schema['types']):
        if has_fields(tdef.BaseType):
            for fdef in tdef.Fields:
                # If FieldOptions contains a type option, create an explicit type
                if fto := ftopts_s2d(fdef.FieldOptions)[1]:
                    # Move all type options to new type
                    newopts = [fdef.FieldOptions.pop(get_optx(fdef.FieldOptions, o)) for o in fto]
                    name = epname(newopts, sys)              # If enum/pointer option, use derived enum typename
                    newname = name if name else f'{tdef.TypeName}{sys}{fdef.FieldName}'
                    if newname not in [t.TypeName for t in schema['types']]:
                        newtype = 'Enumerated' if epx(newopts) is not None else fdef.FieldType
                        assert is_builtin(newtype)      # Don't create a bad type definition
                        schema['types'].append(TypeDefinition(newname, newtype, newopts, fdef.FieldDesc))
                    fdef.FieldType = newname           # Redirect field to explicit type definition


# Generate Enumerated list of fields or JSON Pointers
def simplify_derived_enum(schema: dict, sys: str = '$') -> NoReturn:
    typex = {t[TypeName]: n for n, t in enumerate(schema['types'])}       # Build type index

    def update_eref(enums: dict, opts: List[OPTION_TYPES], optname: str) -> NoReturn:
        n = get_optx(opts, optname)
        if n is not None:
            if name := epname([opts[n][1:]], sys):
                if name in enums:       # Reference existing Enumerated type
                    opts[n] = f'{opts[n][:1]}{enums[name]}'
                else:                   # Make new Enumerated type
                    make_items = enum_items if opts[n][1:2] == OPTION_ID['enum'] else pointer_items
                    opts[n] = f'{opts[n][:1]}{name}'
                    schema['types'].append(TypeDefinition(name, 'Enumerated', [], '', [EnumFieldDefinition(*f) for f in make_items(name.rsplit(sys, maxsplit=1)[0])]))

    def enum_items(rtype: str) -> list:
        tdef = schema['types'][typex[rtype]]
        fields = tdef.Fields if has_fields(tdef.BaseType) else []
        return [[f.FieldID, f.FieldName, f.FieldDesc] for f in fields]

    def pointer_items(rtype: str) -> list:
        def pathnames(rtype: str, base='') -> Generator[list, None, None]:  # Walk subfields of referenced type
            tdef = schema['types'][typex[rtype]]  # TODO: proper error handling for built-in or non-existing reference
            if has_fields(tdef.BaseType):
                for f in tdef.Fields:
                    if OPTION_ID['dir'] in f.FieldOptions:
                        yield from pathnames(f.FieldType, f'{f.FieldName}/')
                    else:
                        yield [base + f.FieldName, f.FieldDesc]
        return [[n+1] + f for n, f in enumerate(pathnames(rtype))]

    enums = {}
    for tdef in list(schema['types']):  # Replace enum/pointer options in Enumerated types with explicit items
        if tdef.BaseType == 'Enumerated':
            to = tdef.TypeOptions
            if rname := epname(to, sys):
                optx = get_optx(to, 'enum')
                optx = optx if optx is not None else get_optx(to, 'pointer')
                items = enum_items if to[optx][:1] == OPTION_ID['enum'] else pointer_items
                tdef.Fields = items(to[optx][1:])
                del to[optx]
                enums.update({rname: tdef.TypeName})

    # Create new Enumerated enum/pointer types if they don't already exist
    for tdef in list(schema['types']):
        if tdef.BaseType in ('ArrayOf', 'MapOf'):
            update_eref(enums, tdef.TypeOptions, 'vtype')
            update_eref(enums, tdef.TypeOptions, 'ktype')


def simplify_map_of_enum(schema: dict) -> NoReturn:
    """
    Replace MapOf(enumerated key) with explicit Map
    """
    typex = {t[TypeName]: n for n, t in enumerate(schema['types'])}       # Build type index
    for n, tdef in enumerate(schema['types']):
        to = topts_s2d(tdef.TypeOptions)
        if tdef.BaseType == 'MapOf' and schema['types'][typex[to['ktype']]][BaseType] == 'Enumerated':
            newfields = [GenFieldDefinition(f[ItemID], f[ItemValue], to['vtype'], [], f[ItemDesc]) for f in schema['types'][typex[to['ktype']]][Fields]]
            schema['types'][n] = TypeDefinition(tdef.TypeName, 'Map', [], tdef.TypeDesc, newfields)


def simplify(schema: dict, extensions: Set[str] = None) -> dict:  # Remove schema extensions
    """
    Given an input schema, return a simplified schema with some or all extensions removed.

    extensions = set of extension names to process:
        AnonymousType:   Replace all anonymous type definitions with explicit
        Multiplicity:    Replace all multi-value fields with explicit ArrayOf type definitions
        DerivedEnum:     Replace all derived and pointer enumerations with explicit Enumerated type definitions
        MapOfEnum:       Replace all MapOf types with listed keys with explicit Map type definitions
    """
    extensions = extensions or EXTENSIONS
    assert extensions - EXTENSIONS == set()
    sys = '$'  # Character reserved for tool-generated TypeNames
    sc = object_type_schema(copy.deepcopy(schema))  # Don't modify original schema

    if 'Multiplicity' in extensions:            # Expand repeated types into ArrayOf defintions
        simplify_multiplicity(sc, sys)
    if 'AnonymousType' in extensions:           # Expand inline definitions into named type definitions
        simplify_anonymous_types(sc, sys)
    if 'DerivedEnum' in extensions:             # Generate Enumerated list of fields or JSON Pointers
        simplify_derived_enum(sc, sys)
    if 'MapOfEnum' in extensions:               # Generate explicit Map from MapOf
        simplify_map_of_enum(sc)
    return list_type_schema(sc)


def get_enum_items(tdef: list, topts: dict, types: dict) -> list:
    def ptr(fdef: list):
        if OPTION_ID['dir'] in fdef[FieldOptions]:
            return f'{fdef[FieldName]}^'
        return fdef[FieldName]

    if 'enum' in topts:
        return types[topts['enum']][Fields]
    if 'pointer' in topts:
        return [(n, ptr(f), '') for n, f in enumerate(types[topts['pointer']][Fields])]
    return tdef[Fields]


__all__ = [
    'get_enum_items',
    'strip_comments',
    'simplify'
]



"""
Load, validate, prettyprint, and dump JSON Abstract Encoding Notation (JADN) schemas
"""
import json
import jsonschema
import numbers
import os
import jadn

from datetime import datetime
from typing import Any, NoReturn, Union
from .definitions import (
    FieldID, FieldName, FieldDesc, FIELD_LENGTH,
    OPTION_ID, REQUIRED_TYPE_OPTIONS, ALLOWED_TYPE_OPTIONS, VALID_FORMATS, is_builtin, has_fields
)
from .utils import raise_error, list_get_default, TypeDefinition, GenFieldDefinition


def data_dir() -> str:
    """
    Return directory containing JADN schema files
    """
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')


# Check schema is valid
def check_typeopts(type_name: str, base_type: str, topts: dict) -> NoReturn:
    """
    Check for invalid type options and undefined formats
    """
    topts_set = set(topts)

    if ro := set(REQUIRED_TYPE_OPTIONS[base_type]) - topts_set:
        raise_error(f'Missing type option {type_name}: {ro}')
    if uo := topts_set - set(ALLOWED_TYPE_OPTIONS[base_type]):
        raise_error(f'Unsupported type option {type_name} ({base_type}): {uo}')
    if 'maxv' in topts and 'minv' in topts and topts['maxv'] < topts['minv']:
        raise_error(f'Bad value range {type_name} ({base_type}): [{topts["minv"]}..{topts["maxv"]}]')
    if 'maxf' in topts and 'minf' in topts and topts['maxf'] < topts['minf']:
        raise_error(f'Bad value range {type_name} ({base_type}): [{topts["minf"]}..{topts["maxf"]}]')

    # TODO: if format defines array, add minv/maxv (prevents adding default max)
    if fmt := topts.get('format'):
        if fmt not in VALID_FORMATS or base_type != VALID_FORMATS[fmt]:
            raise_error(f'Unsupported format {fmt} in {type_name} {base_type}')
    if 'enum' in topts and 'pointer' in topts:
        raise_error(f'Type cannot be both Enum and Pointer {type_name} {base_type}')
    if 'and' in topts and 'or' in topts:
        raise_error(f'Unsupported union+intersection in {type_name} {base_type}')


# TODO: finish convert to use dataclasses??
def check(schema: dict) -> dict:
    """
    Validate JADN schema against JSON schema,
    Validate JADN schema against JADN meta-schema, then
    Perform additional checks on type definitions
    """
    # Add empty Fields if not present
    schema_types = [TypeDefinition(*t) for t in schema['types']]
    schema['types'] = [list(t) for t in schema_types]

    data_path = data_dir()
    with open(os.path.join(data_path, 'jadn_v1.0_schema.json')) as f:     # Check using JSON Schema for JADN
        jsonschema.Draft7Validator(json.load(f)).validate(schema)

    with open(os.path.join(data_path, 'jadn_v1.0_schema.jadn')) as f:     # Optional: check using JADN meta-schema
        meta_schema = jadn.codec.Codec(json.load(f), verbose_rec=True, verbose_str=True, config=schema)
        assert meta_schema.encode('Schema', schema) == schema

    # Additional checks not included in schema
    types = set()
    for type_def in schema_types:
        if type_def.TypeName in types:
            raise_error(f'Duplicate type definition {type_def.TypeName}')
        types.add(type_def.TypeName)
        if is_builtin(type_def.TypeName):
            raise_error(f'Reserved type name {type_def.TypeName}')
        if not is_builtin(type_def.BaseType):
            raise_error(f'Invalid base type {type_def.TypeName}: {type_def.BaseType}')
        type_opts = jadn.topts_s2d(type_def.TypeOptions)
        check_typeopts(type_def.TypeName, type_def.BaseType, type_opts)

        # Check fields
        fields = type_def.Fields
        # Defined fields if there shouldn't be any
        if ('enum' in type_opts or 'pointer' in type_opts) and fields:
            raise_error(f'{type_def.TypeName}({type_def.BaseType}) should not have defined fields with the option enum/pointer')

        # Duplicates
        def duplicates(seq):
            seen = set()
            return set(x for x in seq if x in seen or seen.add(x))

        if dd := duplicates((f[FieldID] for f in fields)):
            raise_error(f'Duplicate fieldID: {type_def.TypeName} {dd}')
        if dd := duplicates((f[FieldName] for f in fields)):
            raise_error(f'Duplicate field name {type_def.TypeName} {dd}')
        # fids = {f[FieldID] for f in fields}  # Field IDs
        # fnames = {f[FieldName] for f in fields}  # Field Names
        # if len(fields) != len(fids) or len(fields) != len(fnames):
        #    raise_error(f'Duplicate field {type_def.TypeName} {len(fields)} fields, {len(fids)} unique tags, {len(fnames)} unique names')

        # Invalid definitions of field
        flen = FIELD_LENGTH[type_def.BaseType]  # Field item count
        if invalid := list_get_default([f for f in fields if len(f) != flen], 0):
            raise_error(f'Bad field id=`{invalid[FieldID]}` in {type_def.TypeName} length, {len(invalid)} should be {flen}')

        # Specific checks
        # Ordinal indexes
        if type_def.BaseType in ('Array', 'Record'):
            if invalid := list_get_default([(f, n) for n, f in enumerate(fields, 1) if f[FieldID] != n], 0):
                field, idx = invalid
                raise_error(f'Item tag error: {type_def.TypeName}({type_def.BaseType}) [{field[FieldName]}] -- {field[FieldID]} should be {idx}')

        # Full Fields -> Array, Choice, Map, Record
        if flen > FieldDesc:  # Full field, not an Enumerated item
            for field in [f if isinstance(f, GenFieldDefinition) else GenFieldDefinition(*f) for f in fields]:
                fo, fto = jadn.ftopts_s2d(field.FieldOptions)
                minc = fo.get('minc', 1)
                maxc = fo.get('maxc', 1)
                if minc < 0 or maxc < 0 or (0 < maxc < minc):
                    raise_error(f'{type_def.TypeName}/{field.FieldName} bad multiplicity {minc} {maxc}')

                if tf := fo.get('tagid', None):
                    if tf not in {f[FieldID] for f in fields}:
                        raise_error(f'{type_def.TypeName}/{field.FieldName}({field.FieldType}) choice has bad external tag {tf}')
                if is_builtin(field.FieldType):
                    check_typeopts(f'{type_def.TypeName}/{field.FieldName}', field.FieldType, fto)
                elif fto:
                    # unique option will be moved to generated ArrayOf
                    allowed = {'unique', } if maxc != 1 else set()
                    if set(fto) - allowed:
                        raise_error(f'{type_def.TypeName}/{field.FieldName}({field.FieldType}) cannot have Type options {fto}')
                if 'dir' in fo:
                    if is_builtin(field.FieldType) and not has_fields(field.FieldType):  # TODO: check defined type
                        raise_error(f'{type_def.TypeName}/{field.FieldName}: {field.FieldType} cannot be dir')
    return schema


def analyze(schema: dict) -> dict:
    def ns(name: str, nsids: dict) -> str:  # Return namespace if name has a known namespace, otherwise return full name
        nsp = name.split(':')[0]
        return nsp if nsp in nsids else name

    # TODO: Check for extension usages
    items = jadn.build_deps(schema)
    # out, roots = topo_sort(items)
    info = schema.get('info', {})
    imports = info.get('imports', {})
    exports = info.get('exports', [])

    defs = set(items) | set(imports)
    refs = {ns(r, imports) for i in items for r in items[i]} | set(exports)
    oids = (OPTION_ID['enum'], OPTION_ID['pointer'])
    refs = {r[1:] if r[0] in oids else r for r in refs}         # Reference base type for derived enums/pointers
    return {
        'unreferenced': list(map(str, defs - refs)),
        'undefined': list(map(str, refs - defs)),
        'cycles': [],
    }


def loads(jadn_str: str) -> dict:
    return check(json.loads(jadn_str))


def load(fname: Union[str, bytes, int]) -> dict:
    with open(fname, encoding='utf-8') as f:
        return check(json.load(f))


def dumps_rec(val: Any, level: int = 0, indent: int = 1, strip: bool = False, nlevel: int = None) -> str:
    if isinstance(val, (numbers.Number, type(''))):
        return json.dumps(val)

    sp2 = (level + 1) * indent * ' '
    sep2 = ',\n' if strip else ',\n\n'
    if isinstance(val, dict):
        sp = level * indent * ' '
        sep = ',\n' if level > 0 else sep2
        lines = sep.join(f'{sp2}"{k}": {dumps_rec(val[k], level + 1, indent, strip)}' for k in val)
        return f'{{\n{lines}\n{sp}}}'
    if isinstance(val, list):
        sep = ',\n' if level > 1 else sep2
        nest = val and isinstance(val[0], list)  # Not an empty list
        vals = [f"{sp2 if nest else ''}{dumps_rec(v, level + 1, indent, strip, level)}" for v in val]
        if nest:
            spn = (nlevel if nlevel else level) * indent * ' '
            return f"[\n{sep.join(vals)}\n{spn}]"
        return f"[{', '.join(vals)}]"
    return '???'


def dumps(schema: dict, strip: bool = False) -> str:
    return dumps_rec(schema, strip=strip)


def dump(schema: dict, fname: Union[str, bytes, int], source: str = '', strip: bool = False) -> NoReturn:
    with open(fname, 'w') as f:
        if source:
            f.write(f'"Generated from {source}, {datetime.ctime(datetime.now())}"\n\n')
        f.write(dumps(schema, strip=strip) + '\n')


__all__ = [
    'analyze',
    'check',
    'dump',
    'dumps',
    'load',
    'loads',
    'data_dir'
]



import re
import jsonschema

from copy import deepcopy
from typing import Any, Dict, Callable
from ..definitions import FORMAT_JS_VALIDATE, FORMAT_VALIDATE, FORMAT_SERIALIZE

ValidationFunction = Callable[[Any], Any]
FormatTable = Dict[str, Dict[str, ValidationFunction]]


def _format_ok(val: Any) -> bool:
    return True


# Return validation functions for the specified format keyword and data type
def get_format_validate_function(format_table: FormatTable, base_type: str, format_kw: str) -> ValidationFunction:
    if not format_kw:
        return _format_ok
    try:
        return format_table[base_type][format_kw]
    except KeyError:
        if format_kw in FORMAT_SERIALIZE:
            return _format_ok           # no value constraints on this keyword
        raise


# Regex from https://stackoverflow.com/questions/201323/how-to-validate-an-email-address-using-a-regular-expression
#   A more comprehensive email address validator is available at http://isemail.info/about
def s_email(sval: str) -> str:
    if not isinstance(sval, type('')):
        raise TypeError
    rfc5322_re = (
        r"(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"
        r'"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@'
        r"(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])"
        r"|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]"
        r":(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])")
    if re.match(rfc5322_re, sval):
        return sval
    raise ValueError


# From https://stackoverflow.com/questions/2532053/validate-a-hostname-string
def s_hostname(sval: str) -> str:
    if not isinstance(sval, type('')):
        raise TypeError
    hostname = sval[:]      # Copy since we're modifying input
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    if len(sval) > 253:
        raise ValueError
    allowed = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    if all(allowed.match(x) for x in hostname.split(".")):
        return sval
    raise ValueError


def val_binary(bval: bytes, condition: Callable) -> bytes:
    if not isinstance(bval, bytes):
        raise TypeError
    if condition(bval):
        return bval
    raise ValueError


def b_mac_addr(bval: bytes) -> bytes:       # Length of MAC addr must be 48 or 64 bits
    return val_binary(bval, lambda x: len(x) == 6 or len(x) == 8)


def b_ipv4_addr(bval: bytes) -> bytes:      # IPv4 address
    return val_binary(bval, lambda x: len(x) == 4)


def b_ipv6_addr(bval: bytes) -> bytes:      # IPv4 address
    return val_binary(bval, lambda x: len(x) == 16)


def _ipnet(aval: [bytes, int], condition: Callable[[list], bool]) -> [bytes, int]:
    if not (isinstance(aval, list) and len(aval) == 2 and isinstance(aval[0], bytes) and isinstance(aval[1], int)):
        raise TypeError
    if condition(aval):
        return aval
    raise ValueError


def a_ipv4_net(aval: [bytes, int]) -> [bytes, int]:       # IPv4 address and netmask
    return _ipnet(aval, lambda x: len(x[0]) == 4 and 0 <= x[1] <= 32)


def a_ipv6_net(aval: [bytes, int]) -> [bytes, int]:       # IPv6 address and netmask
    return _ipnet(aval, lambda x: len(x[0]) == 16 and 0 <= x[1] <= 128)


def val_int(ival: int, condition: Callable[[int], bool]) -> int:
    if not isinstance(ival, int):
        raise TypeError
    if condition(ival):
        return ival
    raise ValueError


def i_i8(ival: int) -> int:
    return val_int(ival, lambda x: -2**7 <= x < 2**7)


def i_i16(ival: int) -> int:
    return val_int(ival, lambda x: -2**15 <= x < 2**15)


def i_i32(ival: int) -> int:
    return val_int(ival, lambda x: -2**31 <= x < 2**31)


def i_i64(ival: int) -> int:
    return val_int(ival, lambda x: -2**63 <= x < 2**63)


# Semantic validation functions
FORMAT_VALIDATE_FUNCTIONS = {
    'String': {
        'email': s_email,
        'hostname': s_hostname,
    },
    'Binary': {
        'eui': b_mac_addr,
        'ipv4-addr': b_ipv4_addr,
        'ipv6-addr': b_ipv6_addr,
    },
    'Array': {
        'ipv4-net': a_ipv4_net,
        'ipv6-net': a_ipv6_net,
    },
    'Integer': {
        'i8': i_i8,
        'i16': i_i16,
        'i32': i_i32,
        'i64': i_i64,
    }
}

# Don't need special format functions, use value constraints:
#   Date-Time       - Integer - min and max value for plausible date range
#   Duration        - Integer - min 0, max value for plausible durations
#   Identifier      - String - regex pattern
#   Port            - Integer - min 0, max 65535


# Create a table of validation functions for format keywords
def format_validators() -> FormatTable:  # Generate validation function table
    # Create a closure for a JSON Schema format keyword
    def make_jsonschema_validator(format_kw: str) -> Callable[[str], str]:
        def validate(val: str) -> str:
            try:
                jsonschema.validate(
                    instance=val,
                    schema={'type': 'string', 'format': format_kw},
                    format_checker=jsonschema.draft7_format_checker
                )
            except jsonschema.exceptions.ValidationError as e:
                raise ValueError(e.message)
            return val
        return validate

    # Ensure code is in sync with jadn_defs
    assert set(FORMAT_VALIDATE) == {y for x in FORMAT_VALIDATE_FUNCTIONS.values() for y in x}

    # Add JSON Schema validation functions to those defined here
    validation_functions = deepcopy(FORMAT_VALIDATE_FUNCTIONS)
    validation_functions['String'].update({f: make_jsonschema_validator(f) for f in FORMAT_JS_VALIDATE})
    return validation_functions



"""
Conversion utilities
"""
from typing import Callable, Tuple, Union
from lxml.etree import Element, tostring  # pylint: disable=E0611
from lxml.html import builder


class Doc:
    """
    Base class generating html/xml documents using context managers
    """
    # Context vars
    parent: Union[None, 'Doc.Tag']
    # Class vars
    init: str
    value: None

    def __init__(self, init: str = None, **kwargs):
        # Context vars
        self.parent = None
        # Class Vars
        self.init = init or ''
        self.value = None

    def getvalue(self, pretty: bool = False) -> str:
        return tostring(self.value, pretty_print=pretty, doctype=self.init).decode()

    def context(self) -> Tuple['Doc', Callable]:
        return self, self.tag

    def tag(self, name: str, text: str = None, **kwargs) -> 'Doc.Tag':
        tmp = self.__class__.Tag(self, name, text, **kwargs)
        (self.parent or self).value.append(tmp.value)
        return tmp

    class Tag:
        """
        Base class for html/xml elements using context managers
        """
        # Context vars
        doc: 'Doc'
        parent: Union[None, 'Tag']
        # Class vars
        value: None

        def __init__(self, doc: 'Doc', name: str, text: str = None, **kwargs):
            # Context vars
            self.doc = doc
            self.parent = None
            # Class Vars
            self.value = None

        def __enter__(self):
            self.parent = self.doc.parent
            self.doc.parent = self

        def __exit__(self, exc_type, exc_val, exc_tb):
            (self.parent or self.doc).value.append(self.value)
            self.doc.parent = self.parent


class DocHTML(Doc):
    """
    class generating html documents using context managers
    """
    # Context vars
    parent: Union[None, 'Tag']
    # Class vars
    init: str
    value: builder.HTML

    def __init__(self, init: str = None, **kwargs):
        super().__init__(init, **kwargs)
        self.value = builder.HTML(**kwargs)

    def getvalue(self, pretty: bool = False) -> str:
        args = {'doctype': self.init} if self.init else {}
        return tostring(self.value, method='html', pretty_print=pretty, **args).decode()

    class Tag(Doc.Tag):
        # Context vars
        doc: 'DocHTML'
        parent: Union[None, 'Tag']
        # Class vars
        value: builder.E

        def __init__(self, doc: 'DocHTML', name: str, text: str = None, **kwargs):
            super().__init__(doc, name, text, **kwargs)
            if cls := kwargs.pop('klass', None):
                kwargs['class'] = cls
            child = (text, ) if text else ()
            self.value = getattr(builder, name.upper())(*child, **kwargs)


class DocXML(Doc):
    """
    Base class generating xml documents using context managers
    """
    # Context vars
    parent: Union[None, 'DocXML.Tag']
    # Class vars
    init: str
    value: Element

    def __init__(self, init: str = None, **kwargs):
        super().__init__(init, **kwargs)
        self.value = builder.HTML(**kwargs)

    class Tag(Doc.Tag):
        # Context vars
        doc: 'DocXML'
        parent: Union[None, 'Tag']
        # Class vars
        value: Element

        def __init__(self, doc: 'DocXML', name: str, text: str = None, **kwargs):
            super().__init__(doc, name, text, **kwargs)
            if cls := kwargs.pop('klass', None):
                kwargs['class'] = cls
            self.value = Element(name, attrib=kwargs)
            if text:
                self.value.text = text

        def __enter__(self):
            self.parent = self.doc.parent
            self.doc.parent = self

        def __exit__(self, exc_type, exc_val, exc_tb):
            (self.parent or self.doc).value.append(self.value)
            self.doc.parent = self.parent


__all__ = [
    'DocHTML',
    'DocXML'
]
